// fbo.h
// PRACA S FRAMEBUFFER OBJEKTAMI, RENDEROVANIE DO TEXTUR

// pouzivame Assimp kniznicu, http://assimp.sourceforge.net/
#include "assimp/quaternion.inl"
#include "assimp/vector3.h"
#include "assimp/matrix4x4.inl"
#include <vector>
#include "glew.h"


class FBO
{
public:
    FBO()
    {
        id = 0;
        color_textures[0] = color_textures[1] = color_textures[2] = color_textures[3] = 0;
        depth_texture = 0;
    };

    // OpenGL ID framebuffer objektu
    GLuint id;
    // id textur pripojenych k FBO, zatial napevno 4
    GLuint color_textures[4];
    // id textury pre depth buffer
    GLuint depth_texture;

    // vymaz framebuffer a jeho textury z OpenGL
    void Clear()
    {
        if (depth_texture > 0)
            glDeleteTextures(1, &depth_texture);
        if (color_textures[0] > 0)
            glDeleteTextures(1, &(color_textures[0]));
        if (color_textures[1] > 0)
            glDeleteTextures(1, &(color_textures[1]));
        if (color_textures[2] > 0)
            glDeleteTextures(1, &(color_textures[2]));
        if (color_textures[3] > 0)
            glDeleteTextures(1, &(color_textures[3]));
        if (id > 0)
            glDeleteFramebuffersEXT(1, &id);
        id = 0;
        color_textures[0] = color_textures[1] = color_textures[2] = color_textures[3] = 0;
        depth_texture = 0;
    }
};

// vytvor a inicializuj zakladne fbo s niekolkymi color buframi a depth bufrom, mozme renderovat aj do floatovych textur
FBO CreateFBO(int width, int height, int num_color_buffers, bool depth_buffer, bool float_color_buffers);

// nastav view a projekcnu maticu na vyrenderovanie sceny spoza zrkadla, zrj=kadlo je dane ako stvoruholnik vo svetovych suradniciach
void SetMirrorMatrices(aiVector3D camera_pos, std::vector<aiVector3D> mirror_points);

// vykresli polygon zrkadla s namapovanou texturou zrkadla
void RenderMirror(std::vector<aiVector3D> mirror_points, GLdouble mirror_view_matrix[16], GLdouble mirror_projection_matrix[16], GLuint mirror_tex);