// geometry.cpp
// DEFINOVANIE ZAKLADNYCH GEOMETRICKYCH OBJEKTOV

#include "geometry.h"
#include <math.h>
#include <memory.h>


// id cube VBO
GLuint g_uiVBOCubeCoords = 0, g_uiVBOCubeTexCoords = 0, g_uiVBOCubeNormals = 0, g_uiVBOCubeTangents = 0, g_uiVBOCubeIndices = 0;

// id sphere VBO
GLuint g_uiVBOSphereCoords = 0, g_uiVBOSphereTexCoords = 0, g_uiVBOSphereNormals = 0, g_uiVBOSphereTangents = 0, g_uiVBOSphereIndices = 0;
GLuint g_uiVBOSphereNumIndices;

// id torus VBO
GLuint g_uiVBOTorusCoords = 0, g_uiVBOTorusTexCoords = 0, g_uiVBOTorusNormals = 0, g_uiVBOTorusTangents = 0, g_uiVBOTorusIndices = 0;
GLuint g_uiVBOTorusNumIndices;

#define PI 3.14159f
#define NULL 0

GLfloat cube_vertices[] = {1,1,1,   -1,1,1,   -1,-1,1,  1,-1,1,  
                          1,1,1,    1,-1,1,   1,-1,-1,  1,1,-1,
                          1,1,1,    1,1,-1,   -1,1,-1,  -1,1,1,
                          -1,1,1,   -1,1,-1,  -1,-1,-1, -1,-1,1,
                          -1,-1,-1, 1,-1,-1,  1,-1,1,   -1,-1,1,
                          1,-1,-1,  -1,-1,-1, -1,1,-1,  1,1,-1 };
GLfloat cube_normals[] = {0,0,1,    0,0,1,   0,0,1,   0,0,1,  
                          1,0,0,    1,0,0,   1,0,0,   1,0,0,
                          0,1,0,    0,1,0,   0,1,0,   0,1,0,
                          -1,0,0,   -1,0,0,  -1,0,0,  -1,0,0,
                          0,-1,0,   0,-1,0,  0,-1,0,  0,-1,0,
                          0,0,-1,   0,0,-1,  0,0,-1,  0,0,-1};
GLfloat cube_tangents[] = {-1,0,0,1,    -1,0,0,1,   -1,0,0,1,  -1,0,0,1,
                           0,-1,0,1,    0,-1,0,1,   0,-1,0,1,  0,-1,0,1,
                           0,0,-1,1,    0,0,-1,1,   0,0,-1,1,  0,0,-1,1,
                           0,0,-1,1,    0,0,-1,1,   0,0,-1,1,  0,0,-1,1,
                           1,0,0,1,     1,0,0,1,    1,0,0,1,   1,0,0,1,
                           -1,0,0,1,    -1,0,0,1,   -1,0,0,1,  -1,0,0,1};
GLfloat cube_texcoords[] = {0,0,   1,0,  1,1,   0,1,  
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1,
                           0,0,   1,0,  1,1,   0,1};
GLuint cube_indices[] = {0,1,2,3,  4,5,6,7,  8,9,10,11,   12,13,14,15,   16,17,18,19,  20,21,22,23};


// priprava VBO pre renderovanie kocky
Model* CreateCubeModel()
{
    // vytvor mesh s geometriou kocky
    Mesh cube_mesh;
    cube_mesh.num_face_indices = 4;
    cube_mesh.num_faces = 6;
    cube_mesh.id_material = -1;

    glGenBuffers(1, &cube_mesh.id_vertices_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, cube_mesh.id_vertices_VBO);
    glBufferData(GL_ARRAY_BUFFER, 24*3*sizeof(float), cube_vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &cube_mesh.id_normals_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, cube_mesh.id_normals_VBO);
    glBufferData(GL_ARRAY_BUFFER, 24*3*sizeof(float), cube_normals, GL_STATIC_DRAW);

    glGenBuffers(1, &cube_mesh.id_tex_coords_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, cube_mesh.id_tex_coords_VBO);
    glBufferData(GL_ARRAY_BUFFER, 24*2*sizeof(float), cube_texcoords, GL_STATIC_DRAW);

    glGenBuffers(1, &cube_mesh.id_tangents_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, cube_mesh.id_tangents_VBO);
    glBufferData(GL_ARRAY_BUFFER, 24*4*sizeof(float), cube_tangents, GL_STATIC_DRAW);
    
    glGenBuffers(1, &cube_mesh.id_indices_VBO);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cube_mesh.id_indices_VBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6*4*sizeof(GLuint), cube_indices, GL_STATIC_DRAW);

    // vyvor jeden node pre kocku
    Node cube_node;
    cube_node.id_meshes.push_back(0);
    memset(cube_node.model_matrix, 0, 16 * sizeof(float));
    cube_node.model_matrix[0] = cube_node.model_matrix[5] = cube_node.model_matrix[10] = cube_node.model_matrix[15] = 1;

    // vytvor strukturu pre model kocky
    Model* result = new Model();
    result->meshes.push_back(cube_mesh);
    result->nodes.push_back(cube_node);
    result->minx = result->miny = result->minz = -1;
    result->maxx = result->maxy = result->maxz = 1;

    return result;
}

// priprava VBO pre renderovanie sfery
Model* CreateSphereModel(int xSlices, int ySlices)
{
    // vytvor mesh pre sferu
    Mesh sphere_mesh;
    sphere_mesh.id_material = -1;
    sphere_mesh.num_face_indices = 4;
    sphere_mesh.num_faces = (xSlices + 0) * (ySlices + 0);

    float* vertices = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* normals = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* texcoords = new float[2 * (xSlices + 1) * (ySlices + 1)];

    for (int i = 0; i <= xSlices; i++)
    {
        float u = i / float(xSlices);
        for (int j = 0; j <= ySlices; j++)
        {
            float v = j / float(ySlices);

            int index = i + j * (xSlices + 1);

            vertices[3 * index + 0] = cos(2 * u * PI) * cos((v - 0.5f) * PI);
            vertices[3 * index + 1] = sin(2 * u * PI) * cos((v - 0.5f) * PI);
            vertices[3 * index + 2] = sin((v - 0.5f) * PI);

            normals[3 * index + 0] = vertices[3 * index + 0];
            normals[3 * index + 1] = vertices[3 * index + 1];
            normals[3 * index + 2] = vertices[3 * index + 2];

            texcoords[2 * index + 0] = u;
            texcoords[2 * index + 1] = 1.0f - v;
        }
    }

    GLuint* indices = new GLuint[4 * (xSlices + 0) * (ySlices + 0)];
    int index = 0;
    for (int i = 0; i < xSlices; i++)
    {
        for (int j = 0; j < ySlices; j++)
        {
            indices[index] = i + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + (j + 1) * (xSlices + 1);
            index++;
            indices[index] = i + (j + 1) * (xSlices + 1);
            index++;
        }
    }

    // napln VBO obsahujuce polia atributov
    glGenBuffers(1, &sphere_mesh.id_vertices_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, sphere_mesh.id_vertices_VBO);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &sphere_mesh.id_normals_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, sphere_mesh.id_normals_VBO);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), normals, GL_STATIC_DRAW);

    glGenBuffers(1, &sphere_mesh.id_tex_coords_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, sphere_mesh.id_tex_coords_VBO);
    glBufferData(GL_ARRAY_BUFFER, 2 * (xSlices + 1) * (ySlices + 1) * sizeof(float), texcoords, GL_STATIC_DRAW);

    // vypocitaj tangenty
    float* tangents = new float[4 * (xSlices + 1) * (ySlices + 1)];
    CalculateTangentArray((xSlices + 1) * (ySlices + 1), vertices, normals, texcoords, GL_QUADS, 
                          g_uiVBOSphereNumIndices, indices, tangents);
    glGenBuffers(1, &sphere_mesh.id_tangents_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, sphere_mesh.id_tangents_VBO);
    glBufferData(GL_ARRAY_BUFFER, 4 * (xSlices + 1) * (ySlices + 1) * sizeof(float), tangents, GL_STATIC_DRAW);

    // pole indexov
    glGenBuffers(1, &sphere_mesh.id_indices_VBO);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sphere_mesh.id_indices_VBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * (xSlices + 0) * (ySlices + 0) * sizeof(GLuint), indices, GL_STATIC_DRAW);

    delete [] vertices;
    delete [] normals;
    delete [] texcoords;
    delete [] indices;
    delete [] tangents;

    // vyvor jeden node pre sferu
    Node sphere_node;
    sphere_node.id_meshes.push_back(0);
    memset(sphere_node.model_matrix, 0, 16 * sizeof(float));
    sphere_node.model_matrix[0] = sphere_node.model_matrix[5] = sphere_node.model_matrix[10] = sphere_node.model_matrix[15] = 1;

    // vytvor strukturu pre model sfery
    Model* result = new Model();
    result->meshes.push_back(sphere_mesh);
    result->nodes.push_back(sphere_node);
    result->minx = result->miny = result->minz = -1;
    result->maxx = result->maxy = result->maxz = 1;

    return result;
}

// priprava VBO pre renderovanie torusu
Model* CreateTorusModel(float innerRadius, float outerRadius, int xSlices, int ySlices)
{
    // vytvor mesh pre torus
    Mesh torus_mesh;
    torus_mesh.id_material = -1;
    torus_mesh.num_face_indices = 4;
    torus_mesh.num_faces = (xSlices + 0) * (ySlices + 0);

    float* vertices = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* normals = new float[3 * (xSlices + 1) * (ySlices + 1)];
    float* texcoords = new float[2 * (xSlices + 1) * (ySlices + 1)];

    for (int i = 0; i <= xSlices; i++)
    {
        float u = i / float(xSlices);
        for (int j = 0; j <= ySlices; j++)
        {
            float v = j / float(ySlices);

            int index = i + j * (xSlices + 1);

            vertices[3 * index + 0] = cos(2 * u * PI) * (outerRadius + innerRadius * cos(2 * v * PI));
            vertices[3 * index + 1] = sin(2 * u * PI) * (outerRadius + innerRadius * cos(2 * v * PI));
            vertices[3 * index + 2] = innerRadius * sin(2 * v * PI);

            float tx = -sin(2 * u * PI);
            float ty = cos(2 * u * PI);
            float tz = 0;
            float sx = cos(2 * u * PI)*(-sin(2 * v * PI));
            float sy = sin(2 * u * PI)*(-sin(2 * v * PI));
            float sz = cos(2 * v * PI);
            normals[3 * index + 0] = ty*sz - tz*sy;
            normals[3 * index + 1] = tz*sx - tx*sz;
            normals[3 * index + 2] = tx*sy - ty*sx;

            texcoords[2 * index + 0] = u;
            texcoords[2 * index + 1] = 1.0f - v;
        }
    }

    GLuint* indices = new GLuint[4 * (xSlices + 0) * (ySlices + 0)];
    int index = 0;
    for (int i = 0; i < xSlices; i++)
    {
        for (int j = 0; j < ySlices; j++)
        {
            indices[index] = i + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + j * (xSlices + 1);
            index++;
            indices[index] = (i + 1) + (j + 1) * (xSlices + 1);
            index++;
            indices[index] = i + (j + 1) * (xSlices + 1);
            index++;
        }
    }

    glGenBuffers(1, &torus_mesh.id_vertices_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, torus_mesh.id_vertices_VBO);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), vertices, GL_STATIC_DRAW);

    glGenBuffers(1, &torus_mesh.id_normals_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, torus_mesh.id_normals_VBO);
    glBufferData(GL_ARRAY_BUFFER, 3 * (xSlices + 1) * (ySlices + 1) * sizeof(float), normals, GL_STATIC_DRAW);

    glGenBuffers(1, &torus_mesh.id_tex_coords_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, torus_mesh.id_tex_coords_VBO);
    glBufferData(GL_ARRAY_BUFFER, 2 * (xSlices + 1) * (ySlices + 1) * sizeof(float), texcoords, GL_STATIC_DRAW);

    // vypocitaj tangenty
    float* tangents = new float[4 * (xSlices + 1) * (ySlices + 1)];
    CalculateTangentArray((xSlices + 1) * (ySlices + 1), vertices, normals, texcoords, GL_QUADS, 
                          g_uiVBOTorusNumIndices, indices, tangents);
    glGenBuffers(1, &torus_mesh.id_tangents_VBO);
    glBindBuffer(GL_ARRAY_BUFFER, torus_mesh.id_tangents_VBO);
    glBufferData(GL_ARRAY_BUFFER, 4 * (xSlices + 1) * (ySlices + 1) * sizeof(float), tangents, GL_STATIC_DRAW);

    // pridaj pole indexov
    glGenBuffers(1, &torus_mesh.id_indices_VBO);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, torus_mesh.id_indices_VBO);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * (xSlices + 0) * (ySlices + 0) * sizeof(GLuint), indices, GL_STATIC_DRAW);

    delete [] vertices;
    delete [] normals;
    delete [] texcoords;
    delete [] indices;
    delete [] tangents;

    // vyvor jeden node pre sferu
    Node torus_node;
    torus_node.id_meshes.push_back(0);
    memset(torus_node.model_matrix, 0, 16 * sizeof(float));
    torus_node.model_matrix[0] = torus_node.model_matrix[5] = torus_node.model_matrix[10] = torus_node.model_matrix[15] = 1;

    // vytvor strukturu pre model sfery
    Model* result = new Model();
    result->meshes.push_back(torus_mesh);
    result->nodes.push_back(torus_node);
    result->minx = result->miny = -(outerRadius + innerRadius);
    result->maxx = result->maxy = outerRadius + innerRadius;
    result->minz = -innerRadius;
    result->maxz = innerRadius;
    
    return result;
}

void CalculateTangentArray(GLuint num_vertices, float* vertices_3D, float* normals_3D,
        float* texcoords_2D, GLenum polygon_mode, int num_indices, GLuint* indices, float* final_tangents_4D)
{
    // kod zalozeny na clanku http://www.terathon.com/code/tangent.html
    // pripravime pomocne polia, ktore budu vo vrcholoch akumulovat tangentu a bitangentu
    float* tangents = new float[3 * num_vertices * 2];
    float* bitangents = tangents + 3 * num_vertices;
    memset(tangents, 0, 3 * sizeof(float) * num_vertices * 2);
    
    int num_polygon_indices = 3;
    if (polygon_mode == GL_QUADS)
        num_polygon_indices = 4;
    for (int polygon = 0; polygon < num_indices / num_polygon_indices; polygon++)
    {
        // ked mame stvoruholnik, musime pocitat 2 trojuholniky, preto tento cyklus
        for (int a = 0; a <= (num_polygon_indices - 3); a++)
        {
            // ziskaj indexy vrcholov aktualneho trojuholnika
            int i1 = indices[num_polygon_indices * polygon + 0];
            int i2 = indices[num_polygon_indices * polygon + a + 1];
            int i3 = indices[num_polygon_indices * polygon + a + 2];
            
            // ziskaj suradnice troch vrcholov trojuholnika
            float v1x = vertices_3D[3 * i1];
            float v1y = vertices_3D[3 * i1 + 1];
            float v1z = vertices_3D[3 * i1 + 2];
            float v2x = vertices_3D[3 * i2];
            float v2y = vertices_3D[3 * i2 + 1];
            float v2z = vertices_3D[3 * i2 + 2];
            float v3x = vertices_3D[3 * i3];
            float v3y = vertices_3D[3 * i3 + 1];
            float v3z = vertices_3D[3 * i3 + 2];
            
            // ziskaj texturne suradnice troch vrcholov trojuholnika
            float w1x = texcoords_2D[2 * i1];
            float w1y = texcoords_2D[2 * i1 + 1];
            float w2x = texcoords_2D[2 * i2];
            float w2y = texcoords_2D[2 * i2 + 1];
            float w3x = texcoords_2D[2 * i3];
            float w3y = texcoords_2D[2 * i3 + 1];
            
            // vypocitaj tangentu a bitangentu
            float x1 = v2x - v1x;
            float x2 = v3x - v1x;
            float y1 = v2y - v1y;
            float y2 = v3y - v1y;
            float z1 = v2z - v1z;
            float z2 = v3z - v1z;
            
            float s1 = w2x - w1x;
            float s2 = w3x - w1x;
            float t1 = w2y - w1y;
            float t2 = w3y - w1y;
            
            float r = 1.0F / (s1 * t2 - s2 * t1);
            float tangent_x = (t2 * x1 - t1 * x2) * r;
            float tangent_y = (t2 * y1 - t1 * y2) * r;
            float tangent_z = (t2 * z1 - t1 * z2) * r;
            float bitangent_x = (s1 * x2 - s2 * x1) * r;
            float bitangent_y = (s1 * y2 - s2 * y1) * r;
            float bitangent_z = (s1 * z2 - s2 * z1) * r;
            
            // naakumuluj vypocitanu tangentu
            tangents[3 * i1 + 0] += tangent_x;
            tangents[3 * i1 + 1] += tangent_y;
            tangents[3 * i1 + 2] += tangent_z;
            tangents[3 * i2 + 0] += tangent_x;
            tangents[3 * i2 + 1] += tangent_y;
            tangents[3 * i2 + 2] += tangent_z;
            tangents[3 * i3 + 0] += tangent_x;
            tangents[3 * i3 + 1] += tangent_y;
            tangents[3 * i3 + 2] += tangent_z;
            
            bitangents[3 * i1 + 0] += bitangent_x;
            bitangents[3 * i1 + 1] += bitangent_y;
            bitangents[3 * i1 + 2] += bitangent_z;
            bitangents[3 * i2 + 0] += bitangent_x;
            bitangents[3 * i2 + 1] += bitangent_y;
            bitangents[3 * i2 + 2] += bitangent_z;
            bitangents[3 * i3 + 0] += bitangent_x;
            bitangents[3 * i3 + 1] += bitangent_y;
            bitangents[3 * i3 + 2] += bitangent_z;
        }
    }
    
    for (int a = 0; a < (int)num_vertices; a++)
    {        
        // pouzijeme Gramov-Schmidtov ortogonalizacny proces, aby sme spravili tangentu kolmu na normalu
        // tu predpokladame, ze tangenta je "skoro" kolma na normalu
        float dot = tangents[3*a+0]*normals_3D[3*a+0] + tangents[3*a+1]*normals_3D[3*a+1] + tangents[3*a+2]*normals_3D[3*a+2];
        final_tangents_4D[4*a+0] = tangents[3*a+0] - dot * normals_3D[3*a+0];
        final_tangents_4D[4*a+1] = tangents[3*a+1] - dot * normals_3D[3*a+1];
        final_tangents_4D[4*a+2] = tangents[3*a+2] - dot * normals_3D[3*a+2];
        // znormalizuj tangentu
        float length = sqrt(final_tangents_4D[4*a+0]*final_tangents_4D[4*a+0]+
                            final_tangents_4D[4*a+1]*final_tangents_4D[4*a+1]+
                            final_tangents_4D[4*a+2]*final_tangents_4D[4*a+2]);
        if (length > 0)
        {
            final_tangents_4D[4*a+0] = final_tangents_4D[4*a+0] / length;
            final_tangents_4D[4*a+1] = final_tangents_4D[4*a+1] / length;
            final_tangents_4D[4*a+2] = final_tangents_4D[4*a+2] / length;
        }
        
        // do stvrtej suradnice tangenty uloz informaciu, ci lokalny tangencialny priestor je pravo alebo lavotocivy
        float cross_x = normals_3D[3*a+1]*final_tangents_4D[4*a+2] - normals_3D[3*a+2]*final_tangents_4D[4*a+1];
        float cross_y = normals_3D[3*a+2]*final_tangents_4D[4*a+0] - normals_3D[3*a+0]*final_tangents_4D[4*a+2];
        float cross_z = normals_3D[3*a+0]*final_tangents_4D[4*a+1] - normals_3D[3*a+1]*final_tangents_4D[4*a+0];
        dot = cross_x * bitangents[3*a+0] + cross_y * bitangents[3*a+1] + cross_z * bitangents[3*a+2];
        final_tangents_4D[4*a+3] = (dot < 0.0f) ? -1.0f : 1.0f;
    }
    
    delete[] tangents;
}

void AddPlaneEdgeIntesection(aiVector3D plane_normal, float plane_d, aiVector3D edge_start, aiVector3D edge_dir, std::vector<aiVector3D>& intersections)
{
    float dot_plane_normal_edge_dir = plane_normal.x * edge_dir.x + plane_normal.y * edge_dir.y + plane_normal.z * edge_dir.z;
    if (dot_plane_normal_edge_dir == 0.0f)
        return;
    float dot_plane_normal_edge_start = plane_normal.x * edge_start.x + plane_normal.y * edge_start.y + plane_normal.z * edge_start.z;
    float t = -(dot_plane_normal_edge_start + plane_d) / dot_plane_normal_edge_dir;
    if (t >= 0.0f && t <= 1.0f) 
        intersections.push_back(aiVector3D(edge_start.x + t*edge_dir.x, edge_start.y + t*edge_dir.y, edge_start.z + t*edge_dir.z));
}

std::vector< std::vector<aiVector3D> > GetCubeSlices(ObjectPose camera_pose, float dist_delta)
{
    // najdi vektor smeru pohladu
    aiMatrix4x4 camera_pose_matrix = GetMatrix(camera_pose);
    aiVector3D dir(-camera_pose_matrix.a3, -camera_pose_matrix.b3, -camera_pose_matrix.c3);

    std::vector< std::vector<aiVector3D> > result;
    for (float d = -2; d < 2; d += dist_delta)
    {
        // najdi prieniky roviny (x * dir.x + y * dir.y + z * dir.z + d = 0) s 12 hranami normalizovanej kockou
        // prieniky na spodnej stene
        std::vector<aiVector3D> edges_intersections;
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,-1,-1), aiVector3D(2,0,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,1,-1), aiVector3D(2,0,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,1,1), aiVector3D(2,0,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,-1,1), aiVector3D(2,0,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(1,-1,-1), aiVector3D(0,2,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,-1,-1), aiVector3D(0,2,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,-1,1), aiVector3D(0,2,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(1,-1,1), aiVector3D(0,2,0), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(1,-1,-1), aiVector3D(0,0,2), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(1,1,-1), aiVector3D(0,0,2), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,1,-1), aiVector3D(0,0,2), edges_intersections);
        AddPlaneEdgeIntesection(dir, d, aiVector3D(-1,-1,-1), aiVector3D(0,0,2), edges_intersections);

        if (edges_intersections.size() < 1)
            continue;

        // usporiadaj najdene prieniky pekne za sebou aby tvorlil polygon
        std::vector<aiVector3D> sorted_edges_intersections;
        sorted_edges_intersections.push_back(edges_intersections[0]);
        edges_intersections.erase(edges_intersections.begin());
        while (edges_intersections.size() > 0)
        {
            aiVector3D current = sorted_edges_intersections[sorted_edges_intersections.size() - 1];
            // najdi medzi priesecnikmi taky, ktory lezi s aktualnym bodom v jednej zo stien kocky
            for (int i = 0; i < (int)edges_intersections.size(); i++)
            {
                if ((current.x == edges_intersections[i].x && abs(current.x) == 1) ||
                    (current.y == edges_intersections[i].y && abs(current.y) == 1) ||
                    (current.z == edges_intersections[i].z && abs(current.z) == 1))
                {
                    sorted_edges_intersections.push_back(edges_intersections[i]);
                    edges_intersections.erase(edges_intersections.begin() + i);
                    break;
                }
            }
        }
        result.push_back(sorted_edges_intersections);
    }

    return result;
}