
// Standardne pre MS win aplikacie
#include <windows.h>

// GLEW kniznica. pre spravu novsich rozsireni
#include "glew.h"
// GL Utility Toolkit (GLUT)
#include "glut.h"

#include <stdio.h>
#include <math.h>
#include <map>

#include "shaders.h"
#include "textures.h"
#include "geometry.h"
#include "fonts.h"
#include "fbo.h"

// Globalne premenne
bool g_is_anim = true;
float g_current_time = 0.0f;
int g_mode = 0;
float g_zoom = 1;
int g_iterations = 50;

// Globalne premenne pre prezeranie sceny
GLdouble g_vzd, g_old_vzd, g_fi, g_xi, g_old_fi, g_old_xi;
GLint g_left_mouse_down_x, g_left_mouse_down_y;
GLint g_right_mouse_down_y;
GLboolean g_right_down = false, g_left_down = false;
GLuint g_window_height, g_window_width;

// id textur
GLuint g_tex_sun, g_tex_lava, g_tex_skybox, g_tex_terrain_diffuse;
GLuint g_vol_tex;

// id shaderov
GLuint g_shaders_volumetric, g_shaders_skybox, g_shaders_phong, g_shaders_mandelbrot;

// modely a svetla
std::vector<ObjectPose> g_light_animation_keys;
std::vector<float> g_light_animation_knots;
std::vector<ObjectPose> g_duck_animation_keys;
std::vector<float> g_duck_animation_knots;
Model* g_duck_model = NULL;
Model* g_car_model = NULL;
Model* g_cube_model = NULL;
Model* g_sphere_model = NULL;
Model* g_torus_model = NULL;

void InitScene()
{
    // nacitaj modely
    g_duck_model = LoadModel("./models/Duck/duck.dae");
    g_car_model = LoadModel("./models/F40/F40.dae");

    // vytvor modely
    g_sphere_model = CreateSphereModel(32, 32);
    g_torus_model = CreateTorusModel(0.6f, 2.0f, 32, 32);
    g_cube_model = CreateCubeModel();

    // animacia svetla
    float dist = 1;
    g_light_animation_knots.push_back(0);
    g_light_animation_keys.push_back(CreatePose(7 * dist, 0 * dist, 3 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(3);
    g_light_animation_keys.push_back(CreatePose(7 * dist, -6 * dist, 4 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(6);
    g_light_animation_keys.push_back(CreatePose(0 * dist, -7 * dist, 5 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(9);
    g_light_animation_keys.push_back(CreatePose(-7 * dist, -6 * dist, 6 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(12);
    g_light_animation_keys.push_back(CreatePose(-7 * dist, 3 * dist, 4 * dist, 0, 0, 0, 1, 1, 1));
    g_light_animation_knots.push_back(15);
    g_light_animation_keys.push_back(CreatePose(7 * dist, 0 * dist, 3 * dist, 0, 0, 0, 1, 1, 1));
}

void InitShaders()
{
    // nacitanie, kompilacia a zlinkovanie shaderov
    g_shaders_volumetric = LoadProgram("shaders/volumetric.vert", NULL, "shaders/volumetric.frag");

    g_shaders_skybox = LoadProgram("shaders/skybox.vert", NULL, "shaders/skybox.frag");

    g_shaders_phong = LoadProgram("shaders/perpixel_phong.vert", NULL, "shaders/perpixel_phong.frag");

    g_shaders_mandelbrot = LoadProgram("shaders/mandelbrot.vert", NULL, "shaders/mandelbrot.frag");
}

// inicializacia OpenGL textur
void InitTextures()
{
    // ************* textury pre teren *************
    g_tex_terrain_diffuse = LoadTexture("textures/terrain_diffuse.png");

    // ************* textura pre slnko *************
    g_tex_sun = LoadTexture("textures/sun.jpg");

    // ************* textura pre skybox *************
    g_tex_skybox = LoadCubeTexture("skybox/snow_positive_x.jpg", "skybox/snow_negative_x.jpg",
                                   "skybox/snow_positive_y.jpg", "skybox/snow_negative_y.jpg",
                                   "skybox/snow_positive_z.jpg", "skybox/snow_negative_z.jpg");

    g_tex_lava = LoadTexture("textures/lava.png");

    std::vector<char*> slices;
    slices.push_back("volume/brain_001.bmp");
    slices.push_back("volume/brain_002.bmp");
    slices.push_back("volume/brain_003.bmp");
    slices.push_back("volume/brain_004.bmp");
    slices.push_back("volume/brain_005.bmp");
    slices.push_back("volume/brain_006.bmp");
    slices.push_back("volume/brain_007.bmp");
    slices.push_back("volume/brain_008.bmp");
    slices.push_back("volume/brain_009.bmp");
    slices.push_back("volume/brain_010.bmp");
    slices.push_back("volume/brain_011.bmp");
    slices.push_back("volume/brain_012.bmp");
    slices.push_back("volume/brain_013.bmp");
    slices.push_back("volume/brain_014.bmp");
    slices.push_back("volume/brain_015.bmp");
    slices.push_back("volume/brain_016.bmp");
    slices.push_back("volume/brain_017.bmp");
    slices.push_back("volume/brain_018.bmp");
    slices.push_back("volume/brain_019.bmp");
    slices.push_back("volume/brain_020.bmp");
    g_vol_tex = Load3DTexture(slices);
}

// Open GL Inicializacia
bool InitGL(void)
{
    // vypis info o systeme OpenGL
    printf("OpenGL version: %s\n", glGetString(GL_VERSION));
    printf("OpenGL vendor: %s\n", glGetString(GL_VENDOR));
    printf("OpenGL renderer: %s\n", glGetString(GL_RENDERER));

    GLint maxColorAttachments;
    glGetIntegerv(GL_MAX_COLOR_ATTACHMENTS_EXT, &maxColorAttachments);
    printf("Max FBO color attachments: %i\n", maxColorAttachments);

    GLfloat maximumAnisotropy;
    glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maximumAnisotropy);
    printf("Max Anisotrphy filter samples: %i\n", int(maximumAnisotropy));

    GLfloat maxTexturesInVertexShader;
    glGetFloatv(GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS, &maxTexturesInVertexShader);
    printf("Max Vertex Shader Textures: %i\n", int(maxTexturesInVertexShader));

    // nastav cierne pozadie
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    // nastav z-buffer testovanie
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // nastav stencil buffer
    glClearStencil(0);

    // nastav zmiesavaciu funkciu
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // nastav vhodne mapovanie textur
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // nastavenie svetla
    GLfloat sun_color[] = {0.9f, 0.95f, 0.9f, 1.0f};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, sun_color);

    // initializuj kameru
    g_fi = 0.7f; g_xi = 0.7f; g_vzd = 10.0f;

	return true;
}

// vyrenderuj celu scenu, modelview je v tomto momente uz nastavena a obsahuje view maticu kamery
void RenderWorldScene()
{
    // vykresli skybox
    // OBJEKTY POUZIVAJUCE SKYBOX SHADER
    // skybox je dany vo svetovych suradniciach, akurat jeho stred je stale v bode, kde je kamera
    glUseProgram(g_shaders_skybox);
    SetShaderUniform1i(g_shaders_skybox, "cube_map", 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_CUBE_MAP, g_tex_skybox);
    glPushMatrix();
    glScalef(100, 100, 100);
    RenderModel(g_cube_model, g_shaders_skybox);
    glPopMatrix();
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

    // vykresli slnko na pozicii svetla
    GLuint current_shader = g_shaders_phong;
    glUseProgram(current_shader);
    SetShaderUniform1i(current_shader, "lighting_enabled", 0);
    SetShaderUniform1i(current_shader, "texturing_enabled", 1);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, g_tex_sun);
    glPushMatrix();
    ObjectPose light_pose = IntepolatePoseCubic(g_light_animation_knots, g_light_animation_keys, g_current_time, false);
    MultiplyMatrixOpenGL(GetMatrix(light_pose));
    GLfloat sun_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    glLightfv(GL_LIGHT0, GL_POSITION, sun_position);
    glScalef(0.2f, 0.2f, 0.2f);
    glColor3f(1.0f, 1.0f, 1.0f);
    RenderModel(g_sphere_model, current_shader);
    glPopMatrix();

    // vykresli obdlznik s namapovanym mandelbrotom
    glUseProgram(g_shaders_mandelbrot);
    GLfloat zoom_rectangle[4];
    zoom_rectangle[2] = zoom_rectangle[3] = 4.0f / g_zoom;
    if (g_mode == 0)
    {
        zoom_rectangle[0] = 0.001643721971153f - zoom_rectangle[2] / 2.0f;
        zoom_rectangle[1] = -0.822467633298876f - zoom_rectangle[3] / 2.0f;
    }
    else
    {
        zoom_rectangle[0] = 0.005f - zoom_rectangle[2] / 2.0f;
        zoom_rectangle[1] = -0.3995f - zoom_rectangle[3] / 2.0f;
    }
    SetShaderUniform4f(g_shaders_mandelbrot, "zoom_rectangle", zoom_rectangle);
    GLfloat julia_param[3];
    julia_param[0] = 0.393f;
    julia_param[1] = -0.34f;
    julia_param[2] = g_mode;
    SetShaderUniform3f(g_shaders_mandelbrot, "julia_param", julia_param);
    GLfloat plane_color[] = { 1.0f, 1.0f, 0.3f, 4.0f };
    SetShaderUniform4f(g_shaders_mandelbrot, "base_color", plane_color);
    SetShaderUniform1i(g_shaders_mandelbrot, "iterations", g_iterations);
    float quad_size = 3;
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0); glVertex3f(-quad_size, -quad_size, 0);
    glTexCoord2f(1, 0); glVertex3f(quad_size, -quad_size, 0);
    glTexCoord2f(1, 1); glVertex3f(quad_size, quad_size, 0);
    glTexCoord2f(0, 1); glVertex3f(-quad_size, quad_size, 0);
    glEnd();

    // vykresli sferu s namapovany fraktalom
    glPushMatrix();
    GLfloat sphere_color[] = { 1.0f, 0.0f, 0.6f, 4.0f };
    SetShaderUniform4f(g_shaders_mandelbrot, "base_color", sphere_color);
    ObjectPose sphere_pose = CreatePose(3, 0, 2, 0, 200*g_current_time, 0, 1, 1, 1);
    MultiplyMatrixOpenGL(GetMatrix(sphere_pose));
    RenderModel(g_sphere_model, g_shaders_mandelbrot);
    glPopMatrix();

    // vykresli sferu s namapovany fraktalom
    glPushMatrix();
    GLfloat torus_color[] = { 0.0f, 1.0f, 0.9f, 4.0f };
    SetShaderUniform4f(g_shaders_mandelbrot, "base_color", torus_color);
    ObjectPose torus_pose = CreatePose(0, 3, 2, 0, 0, 180*g_current_time, 0.3f, 0.4f, 0.3f);
    MultiplyMatrixOpenGL(GetMatrix(torus_pose));
    RenderModel(g_torus_model, g_shaders_mandelbrot);
    glPopMatrix();
}

// Procedura obsluhujuca vykreslovanie
void render(void)   
{
    GLfloat camera_pos[4];
    camera_pos[0] = g_vzd * cos(g_fi) * cos(g_xi);
	camera_pos[1] = g_vzd * sin(g_fi) * cos(g_xi);
	camera_pos[2] = g_vzd * sin(g_xi);
    camera_pos[3] = 1;

    // renderujeme do okna
    // vycistime framebuffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Nastav viewport na velkost okna
    glViewport(0, 0, g_window_width, g_window_height);

    // Nastav projekcnu maticu
    glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();
    gluPerspective(60, g_window_width / float(g_window_height), 0.1, 300.0);

    // Nastav view maticu kamery
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
    ObjectPose camera_pose = CreateLookAtPose(camera_pos[0], camera_pos[1], camera_pos[2], 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f);
    MultiplyMatrixOpenGL(GetMatrix(camera_pose).Inverse());

    // vykresli scenu
    RenderWorldScene();

    glUseProgram(g_shaders_volumetric);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_3D, g_vol_tex);
    SetShaderUniform1i(g_shaders_volumetric, "volumetric_map", 0);
    float deltas[] = {1.0f / 256.0f, 1.0f / 256.0f, 1.0f / 20.0f, 1.0f};
    SetShaderUniform4f(g_shaders_volumetric, "gradient_delta", deltas);
    std::vector< std::vector<aiVector3D> > slices = GetCubeSlices(camera_pose, 0.01f);
    // vykresli za sebou vsetky rezy
    glPushMatrix();
    glTranslatef(0, 0, 2);
    for (int i = 0; i < (int)slices.size(); i++)
    {
        glColor3f(i / float(slices.size()), i / float(slices.size()), 1.0f);
        glBegin(GL_POLYGON);
        for (int j = 0; j < (int)slices[i].size(); j++)
            glVertex3f(slices[i][j].x, slices[i][j].y, slices[i][j].z);
        glEnd();
    }
    glPopMatrix();
    glDisable(GL_BLEND);

    // vykresli 2D text, bez shaderov
    glUseProgram(NULL);
    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_LIGHTING);
    glColor3f(1.0f, 1.0f, 0.0f);
    render2DTextFT(10, g_window_height - 20, "Projekt GRC ...");
    render2DTextFT(10, g_window_height - 45, "F1 - zapni/vypni animaciu");
    if (g_mode == 0)
        render2DTextFT(10, g_window_height - 70, "F2 - mod: MANDELBROT");
    else if (g_mode == 1)
        render2DTextFT(10, g_window_height - 70, "F2 - mod: JULIA");
    render2DTextFT(10, g_window_height - 95, "F3/F4 - zoom in/out fractal");
    char buffer[255];
    sprintf(buffer, "Iteracie: %i (F5/F6 pre zmenu)", g_iterations);
    render2DTextFT(10, g_window_height - 120, buffer);

    // Zamen buffre a zobraz vykreslene objekty
    glutSwapBuffers();
}

// Procedura obsluhujuca zmenu rozmerov okna
void reshape(int w, int h)
{
    g_window_width = w;
    g_window_height = h;
	if (g_window_height == 0) 
        g_window_height = 1;
}

// Procedura obsluhujuca stlacenie klavesy
void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
    {
        // Ak Esc bolo stlacene
		case 27:
            // vymaz alokovane OpenGL data
            ClearModel(g_duck_model);
            ClearModel(g_car_model);
            ClearModel(g_sphere_model);
            ClearModel(g_torus_model);
            ClearModel(g_cube_model);
            ClearProgram(g_shaders_volumetric);
            ClearProgram(g_shaders_skybox);
            ClearTexture(g_tex_sun);
            ClearTexture(g_tex_lava);
            ClearTexture(g_tex_skybox);
            ClearTexture(g_tex_terrain_diffuse);
            // Ukoncenie programu
			exit(0);
		    break;
		default:        
		    break;
	}
}

// Procedura obsluhujuca stlacenie specialnej klavesy
void special_keys(int a_keys, int x, int y)
{
	switch (a_keys) 
    {
        // Ak F1 bolo stlacene
		case GLUT_KEY_F1:
            g_is_anim = !g_is_anim;
            glutPostRedisplay();
		    break;
        // Ak F2 bolo stlacene
        case GLUT_KEY_F2:
            g_mode++;
            if (g_mode > 1)
                g_mode = 0;
            glutPostRedisplay();
		    break;
        case GLUT_KEY_F3:
            g_zoom += 0.2f * g_zoom;
            if (g_zoom > 20000)
                g_zoom = 20000;
            glutPostRedisplay();
		    break;
        case GLUT_KEY_F4:
            g_zoom -= 0.2f * g_zoom;
            if (g_zoom < 1)
                g_zoom = 1;
            glutPostRedisplay();
		    break;
        case GLUT_KEY_F5:
            g_iterations++;
            if (g_iterations > 100)
                g_iterations = 100;
            glutPostRedisplay();
		    break;
        case GLUT_KEY_F6:
            g_iterations--;
            if (g_iterations < 10)
                g_iterations = 10;
            glutPostRedisplay();
		    break;
		default:
		    break;
	}
}

// Procedura obsluhujuca stlacenie tlacitka mysi
void mouse_klik(int button, int state, int x, int y )
{
	switch (button) 
    {
		case GLUT_LEFT_BUTTON:
            if (state == GLUT_DOWN)
            {
                g_left_down = true;
			    g_left_mouse_down_x = x;
			    g_left_mouse_down_y = y;
			    g_old_fi = g_fi;
			    g_old_xi = g_xi;
                break;
            }
            else if (state == GLUT_UP)
            {
                g_left_down = false;
            }
		    break;
        case GLUT_RIGHT_BUTTON:
            if (state == GLUT_DOWN)
            {
                g_right_down = true;
			    g_right_mouse_down_y = y;
			    g_old_vzd = g_vzd;
                break;
            }
            else if (state == GLUT_UP)
            {
                g_right_down = false;
            }
		    break;
		default:
		    break;
	}

    glutPostRedisplay();
}

// Procedura obsluhujuca pohyb mysi
void mouse_move(int x, int y)
{
    if (g_left_down == GL_TRUE)
    {
        g_fi = g_old_fi + (g_left_mouse_down_x - x)/200.0f;
        g_xi = g_old_xi + (y - g_left_mouse_down_y)/200.0f;
        glutPostRedisplay();
    }

    if (g_right_down == GL_TRUE)
    {
        g_vzd = g_old_vzd + (g_right_mouse_down_y - y)/10.0f;
        if (g_vzd > 80) g_vzd = 80;
        glutPostRedisplay();
    } 

}

void timer(int id)
{
    if (g_is_anim)
    {
        g_current_time += 0.050f;
    }
    else
    {
        //g_current_time = 0.0f;
    }

    // jednoduche opakovanie animacie po 17 sekundach
    if (g_current_time > 17)
        g_current_time = 0.0f;

    glutPostRedisplay();
    glutTimerFunc(50, timer, 0);
}

// Zakladna funkcia spustena na zaciatku
int main(int argc, char** argv)
{
    // GLUT Inicializacia
	glutInit(&argc, argv);                              
    // Display Mod (Rgb a Double Buffering)
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
    //glutInitDisplayString("stencil>=8 rgb>=8 double depth>=24");
    // Nastavime pociatocnu poziciu a velkost okna
    glutInitWindowPosition(50, 50);
    glutInitWindowSize(800, 600);                   
    // Vyrobime okno s danym menom
    glutCreateWindow("--== GRC projekt ==--" );  

    // Zinicializujeme GLEW
    glewInit();

    // Nastavime zakladne vlastnosti OpenGL
	InitGL();
    InitTextures();
    InitShaders();
    init2DFontFT("fonts/optien.ttf");
    InitScene();

    // Zaregistrujeme callback funkcie
	glutDisplayFunc(render);                            
	glutReshapeFunc(reshape);                        
	glutKeyboardFunc(keyboard);                      
	glutSpecialFunc(special_keys);
    glutMouseFunc(mouse_klik);
    glutMotionFunc(mouse_move);
    glutTimerFunc(0, timer, 0);

    // Spustenie hlavneho okruhu zachytavania sprav
	glutMainLoop();

	return 0;
}
