// loader.h
// IMPLEMENTACIA NACITANIA MODELOV Z EXTERNEHO SUBORU

// GLEW kniznica. pre spravu novsich rozsireni
#include "glew.h"
#include "pose.h"
#include <vector>

class Mesh
{   
public:
    Mesh()
    {
        id_vertices_VBO = 0;
        id_tex_coords_VBO = 0;
        id_normals_VBO = 0;
        id_indices_VBO = 0;
        id_tangents_VBO = 0;
        num_face_indices = 3;
        num_faces = 0;
        id_material = -1;
    };

    // geometria ulozena vo VBO OpenGL
    GLuint id_vertices_VBO;
    GLuint id_tex_coords_VBO;
    GLuint id_normals_VBO;
    GLuint id_tangents_VBO;
    GLuint id_indices_VBO;
    unsigned int num_faces;

    // pocet vrcholov kazdeho primitivu v meshi
    unsigned int num_face_indices;

    // odkaz na material
    int id_material;
};

class Material
{
public:
    Material()
    {
        id_diffuse_texture = 0;
        color_ambient[0] = color_ambient[1] = color_ambient[2] = 0.2f;
        color_ambient[3] = 1.0f;
        color_diffuse[0] = color_diffuse[1] = color_diffuse[2] = 0.8f;
        color_diffuse[3] = 1.0f;
        color_specular[0] = color_specular[1] = color_specular[2] = 0.0f;
        color_specular[3] = 1.0f;
        color_emission[0] = color_emission[1] = color_emission[2] = 0.0f;
        color_emission[3] = 1.0f;
        shininess = 16;
    };

    // materialove vlastnosti
    GLuint id_diffuse_texture;
    GLuint id_specular_texture;
    GLuint id_normal_texture;
    GLuint id_bump_texture;
    float color_ambient[4];
    float color_diffuse[4];
    float color_specular[4];
    float color_emission[4];
    float shininess;
};

class Node
{
public:
    Node()
    {
        for (int i = 0; i < 16; i++)
            model_matrix[i] = 0.0f;
        model_matrix[0] = model_matrix[5] = model_matrix[10] = model_matrix[15] = 1.0f;
    }

    // model transformacia pre node
    float model_matrix[16];
    std::vector<int> id_meshes;
};

class Model
{
public:
    std::vector<Mesh> meshes;
    std::vector<Material> materials;
    std::vector<Node> nodes;

    // AABB, ohranicujuci objem celeho modelu
    float minx, miny, minz;
    float maxx, maxy, maxz;
};

class SceneNode
{
public:
    Model* model;
    std::vector<ObjectPose> animation_poses;
    std::vector<float> animation_knots;
};

// nacitanie modelu do vnutornych OpenGL struktur z externeho suboru
// podporovane formaty: COLLADA, 3DS, OBJ, ...
// navratova hodnota je zoznam meshov
Model* LoadModel(const char* model_file);

// vykresli mesh aj s materialom
long RenderModel(Model* model, GLuint id_shader, int num_instances = 1);

// vymaz instanciu modelu ako aj vsetky bindnute OpenGL data
void ClearModel(Model* model);

// vykresli node sceny, t.j. model s danou pozou
long RenderSceneNode(SceneNode node, float time, GLuint id_shader, int num_instances = 1);

// vypocitaj ohranicujuci objem nodu na zaklade lokalneho ohranicujuceho modelu a aktualnej pozy
void ComputeAABB(SceneNode node, float time, float& minx, float& miny, float& minz, float& maxx, float& maxy, float& maxz);

// vykresli ohranicujuci objem pre node sceny
void RenderSceneNodeAABB(SceneNode node, float time, bool wireframe);

// zisti ci sa dany bod, ktory je zadany v svetovych suradniciach, nachadza v pohladovom objeme kamery, pricom kamera je dana view a projekcno maticou
bool WorldPointInsideCameraViewVolume(float view_matrix[16], float projection_matrix[16], float pointx, float pointy, float pointz);