// pose.h
// DEFINOVANIE POZY OBJEKTOV, INTERPOLACIA HODNOT

// pouzivame Assimp kniznicu, http://assimp.sourceforge.net/
#include "assimp/quaternion.inl"
#include "assimp/vector3.h"
#include "assimp/matrix4x4.inl"
#include <vector>

class ObjectPose
{
public:
    ObjectPose()
    {
        pos_x = pos_y = pos_z = 0.0f;
        scale_x = scale_y = scale_z = 1.0f;
        rotation = aiQuaternion(0, 0, 0);
    };

    // pozicia objektu
    float pos_x, pos_y, pos_z;
    // rotacia objektu
    aiQuaternion rotation;
    // skalovanie objektu
    float scale_x, scale_y, scale_z;
};

// vytvor pozu na zaklade pozicie, rotacie zadanej eulerovymi uhlami a an zaklade skalovacich faktorov
ObjectPose CreatePose(float pos_x, float pos_y, float pos_z, float euler_x, float euler_y, float euler_z, float scale_x, float scale_y, float scale_z);

// preinterpoluj zadane hodnoty v danom case, pouzije sa linearna interpolacia
float IntepolateValuesLinear(std::vector<float> knots, std::vector<float> values, float time, bool closed);

// preinterpoluj zadane hodnoty v danom case, pouzije sa kubicka interpolacia
float IntepolateValuesCubic(std::vector<float> knots, std::vector<float> values, float time, bool closed);

// preinterpoluj zadane kvaterniony v danom case, pouzije sa linearna sfericka interpolacia
aiQuaternion IntepolateQuaternionsLinear(std::vector<float> knots, std::vector<aiQuaternion> quaternions, float time, bool closed);

// preinterpoluj zadane kvaterniony v danom case, pouzije sa kubicka sfericka interpolacia
aiQuaternion IntepolateQuaternionsCubic(std::vector<float> knots, std::vector<aiQuaternion> quaternions, float time, bool closed);

// preinterpoluj zadane pozy v danom case, pouzije sa linearna interpolacia
ObjectPose IntepolatePoseLinear(std::vector<float> knots, std::vector<ObjectPose> poses, float time, bool closed);

// preinterpoluj zadane pozy v danom case, pouzije sa kubicka interpolacia
ObjectPose IntepolatePoseCubic(std::vector<float> knots, std::vector<ObjectPose> poses, float time, bool closed);

// preinterpoluj zadane pozy v danom case, pouzije sa kubicka interpolacia
ObjectPose IntepolatePoseCubicDerivedRot(std::vector<float> knots, std::vector<ObjectPose> poses, aiQuaternion base, float time, bool closed);

// vytvor maticu danej pozy, vysledna matica predstavuje transformaciu z objektoveko priestoru objektu do svetoveho priestoru
aiMatrix4x4 GetMatrix(ObjectPose pose);

// vytvor assimp maticu z OpenGL matice
aiMatrix4x4 GetMatrixFromOpenGLMatrix(float* opengl_matrix);

// vytvor OpenGL maticu z assimp matice
void GetOpenGLMatrixFromMatrix(aiMatrix4x4 matrix, float* opengl_matrix);

// pridaj danu assimp maticu do aktualnej matice v OpenGL
void MultiplyMatrixOpenGL(aiMatrix4x4 matrix);

// vytvor pozu pri ktorej sa divame z bodu from do bodu to pricom vektor hore je up
// tato poza ma z-tovu suradnicu v smere opacnom ako sa divame, x-ovu v smere doprava od smeru pohladu a y-ovu suradnicu v smere hore od pohladu
ObjectPose CreateLookAtPose(float from_x, float from_y, float from_z, float to_x, float to_y, float to_z, float up_x, float up_y, float up_z);

// vytvor pozu pri ktorej sa divame zo zrkadleneho bodu from do zrkadleneho bodu to pricom zrkadlovy vektor hore je up
// zrkadlenie je dane bodmi, ktore definuju rovinu zrkadlenia
ObjectPose CreateMirrorLookAtPose(float from_x, float from_y, float from_z, float to_x, float to_y, float to_z, float up_x, float up_y, float up_z, std::vector<aiVector3D> mirror_points);

// vytvor OpenGL maticu, ktora predstavuje projekciu, pricom pohladovy objem obsahuje celu scenu danu ohranicujucimi hodnotami
void GetCameraProjectionBasedOnSceneAABB(ObjectPose camera_pose, float scene_min_x, float scene_min_y, float scene_min_z, float scene_max_x, float scene_max_y, float scene_max_z, bool ortho,
                                         float& left, float& right, float& bottom, float& top, float& near, float& far);