
varying vec4 R_world;
varying vec4 V_world;
varying vec4 N_world;
varying vec4 L_world;

uniform sampler2D diffuse_map;
uniform samplerCube cube_map;
uniform vec4 refraction_ratios;

void main(void)
{
   vec4 color_refl = textureCube(cube_map, normalize(R_world.xzy));
   color_refl = mix(color_refl, texture2D(diffuse_map, gl_TexCoord[0].st), 0.5);
   
   vec4 N = normalize(N_world);
   vec4 L = normalize(L_world);
   vec4 V = normalize(V_world);
   
   vec4 TR_world = refract(-V, N, refraction_ratios.r);
   vec4 TG_world = refract(-V, N, refraction_ratios.g);
   vec4 TB_world = refract(-V, N, refraction_ratios.b);
   float refr_r = textureCube(cube_map, TR_world.xzy).r;
   float refr_g = textureCube(cube_map, TG_world.xzy).g;
   float refr_b = textureCube(cube_map, TB_world.xzy).b;
   vec4 color_refr = vec4(refr_r, refr_g, refr_b, 1.0);
   
   float VdotRefL = dot(V, reflect(-L, N));
   vec4 light_reflection = pow(clamp(VdotRefL, 0.0, 1.0), 16.0) * gl_LightSource[0].diffuse;
   
   float fresnel = 1.0 - pow(clamp(dot(N, V), 0.0, 1.0), 5.0);
   
   gl_FragColor = mix(color_refl + light_reflection, color_refr, fresnel);
}
