varying vec2 vTexCoord;

uniform sampler2D map;
uniform sampler2D depth_map;
uniform int samples;
uniform mat4 previous_frame_modelview_matrix;
uniform mat4 previous_frame_projection_matrix;
uniform mat4 current_frame_inverse_modelview_matrix;

void main() 
{
    // reconstruct eye and then object coordinates of fragment
    float A = gl_ProjectionMatrix[0][0];
    float B = gl_ProjectionMatrix[2][0];
    float C = gl_ProjectionMatrix[1][1];
    float D = gl_ProjectionMatrix[2][1];
    float T1 = gl_ProjectionMatrix[2][2];
    float T2 = gl_ProjectionMatrix[3][2];
    float E1 = gl_ProjectionMatrix[2][3];
    float xn = 2.0 * vTexCoord.s - 1.0;
    float yn = 2.0 * vTexCoord.t - 1.0;
    float zn = 2.0 * texture2D(depth_map, vTexCoord).r - 1.0;
    float ze = T2 / (E1 * zn - T1);
    float xe = (E1 * ze * xn - B * ze) / A;
    float ye = (E1 * ze * yn - D * ze) / C;
    vec4 object_space_pos = current_frame_inverse_modelview_matrix * vec4(xe, ye, ze, 1.0);
    
    // get screen space coordinates of this fragment in previous frame
    vec4 clip_coords_prev_frame = previous_frame_projection_matrix * previous_frame_modelview_matrix * object_space_pos;
    vec2 norm_screen_space_prev_frame = vec2(0.5 * clip_coords_prev_frame.x / clip_coords_prev_frame.w + 0.5, 0.5 * clip_coords_prev_frame.y / clip_coords_prev_frame.w + 0.5);

    // our original texcoord for this fragment
	vec2 velocity = 2.5 * (vTexCoord - norm_screen_space_prev_frame);
    vec4 result = texture2D(map, vTexCoord);
    for (int i = 1; i < samples; ++i) 
    {
        vec2 offset = velocity * (float(i) / float(samples - 1) - 0.5);
  
        // sample & add to result:
        result += texture2D(map, vTexCoord + offset);
   }
 
   gl_FragColor = result / float(samples);
}

