varying vec4 V_eye;
varying vec4 L_eye;
varying vec4 N_eye;
varying vec4 shadow_map_tex_coords;

uniform sampler2D diffuse_map;
uniform sampler2D shadow_map;
uniform int texturing_enabled;

void main(void)
{
   vec4 diffuse_material = gl_FrontMaterial.diffuse;
   if (texturing_enabled > 0)
       diffuse_material = texture(diffuse_map, gl_TexCoord[0].st);
   
   vec4 V = normalize(V_eye);
   vec4 L = normalize(L_eye);
   vec4 N = normalize(N_eye); 

   float diffuse = clamp(dot(L, N), 0.0, 1.0);
   vec4 R = reflect(-L, N);
   float specular = sign(diffuse) * pow(clamp(dot(R, V), 0.0, 1.0), gl_FrontMaterial.shininess);

   vec4 depth = texture(shadow_map, vec2(shadow_map_tex_coords));
   if ((depth.r + 0.0001) < shadow_map_tex_coords.z)
   {
       diffuse = 0.3 * diffuse;
	   specular = 0.0;
   }
	   
   vec4 color = 0.2 * (vec4(0.2, 0.2, 0.2, 1.0) + gl_LightSource[0].ambient) * (gl_FrontMaterial.ambient + diffuse_material);
   color += diffuse * gl_LightSource[0].diffuse * diffuse_material;
   color += specular * gl_LightSource[0].specular * gl_FrontMaterial.specular;

   gl_FragColor = color;
}
