varying vec4 V_eye;
varying vec4 L_eye;

uniform sampler3D volumetric_map;
uniform vec4 gradient_delta;

void main(void)
{
   vec4 voxel = texture3D(volumetric_map, gl_TexCoord[0].xyz);

   vec4 sample1, sample2; 
   sample1.x = texture3D(volumetric_map, gl_TexCoord[0].xyz-vec3(gradient_delta.x,0.0,0.0) ).r; 
   sample2.x = texture3D(volumetric_map, gl_TexCoord[0].xyz+vec3(gradient_delta.x,0.0,0.0) ).r; 
   sample1.y = texture3D(volumetric_map, gl_TexCoord[0].xyz-vec3(0.0,gradient_delta.y,0.0) ).r; 
   sample2.y = texture3D(volumetric_map, gl_TexCoord[0].xyz+vec3(0.0,gradient_delta.y,0.0) ).r; 
   sample1.z = texture3D(volumetric_map, gl_TexCoord[0].xyz-vec3(0.0,0.0,gradient_delta.z) ).r; 
   sample2.z = texture3D(volumetric_map, gl_TexCoord[0].xyz+vec3(0.0,0.0,gradient_delta.z) ).r; 
   vec4 N = normalize(sample2 - sample1);
   vec4 V = normalize(V_eye);
   vec4 L = normalize(L_eye); 

   float diffuse = clamp(dot(L, N), 0.0, 1.0);
   vec4 R = reflect(-L, N);
   float specular = sign(diffuse) * pow(clamp(dot(R, V), 0.0, 1.0), 16.0);

   vec4 color = 0.2 * (vec4(0.2, 0.2, 0.2, 1.0) + gl_LightSource[0].ambient) * voxel;
   color += diffuse * gl_LightSource[0].diffuse * voxel;
   color += specular * gl_LightSource[0].specular * vec4(1.0, 1.0, 1.0, 1.0);   

   float alpha = voxel.r;
   if (alpha < 0.3 || alpha > 0.8)
      alpha = 0.0;
   gl_FragColor = vec4(color.rgb, alpha);
}
